/*-
 * Copyright (C)2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: style_copy.c,v 1.3 2020/06/15 17:14:58 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MULAR_COMPAT     0
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

#define PGOBLIN_STYLE_ID "#pGoblin-copy"

typedef struct copy_style {
    char              id[PGOBLIN_STRING_ID_LEN];  /*  "#pGoblin"             */
    u_int32_t         flags;
#   define            PGOBLIN_COPE  0x000100
} copy_style;

static int
/**********************************************************************
 **                                                                  **/
parser(pgoblin_styreg *reg_style, u_char *in) {                     /**
 **                                                                  **
 **********************************************************************/
#   define            ST     ((copy_style *)(reg_style->style))
    int               ex = EX_OK;     /*                                                  */

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!ST) {
        if  (!(reg_style->style = calloc(1, sizeof(copy_style)))) {
            ifBLIN_QX0("No memory #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        strncpy(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID));
        ST->flags = pgoblin_default;
    } else if (!!strncmp(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s", ST->id);
            ERROUT(EX_DATAERR, EINVAL);
    }
    ST->flags &= ~BLIN_MASK;
    for (int i = 0; !!in[i]; ++i) {
        if  (' ' == in[i]) continue;
        if  ('E' == in[i]) ST->flags |= PGOBLIN_COPE;
        break;
    }
out:
    return(ex);
#   undef blin_internal_flags
#   undef  ST
}

static void
/**********************************************************************
 **                                                                  **/
dump(pgoblin_styreg *reg_style) {                                   /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    return;
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
clear(pgoblin_styreg *reg_style) {                                  /**
 **                                                                  **
 **********************************************************************/
    int         ex = EX_OK;
    copy_style *st;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!!reg_style && ((st = reg_style->style))) {
        if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s", st->id);
            ERROUT(EX_DATAERR, EINVAL);
        }
        free(st);
        st = NULL;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static char *
/**********************************************************************
 **                                                                  **/
get(pgoblin_styreg *reg_style, u_char *stringname) {                /**
 **                                                                  **
 **********************************************************************/
    char       *found = NULL;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    return(found);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
tbl( pgoblin_io     *reg_out                                        /**/
   , pgoblin_styreg *reg_style                                      /**/
   , pgoblin_io     *reg_in                                         /**/
   , ssize_t         low                                            /**/
   , ssize_t         hig                                            /**/
   ) {                                                              /**
 **                                                                  **
 **********************************************************************/
    pgoblin_mode   *pq = reg_in->pq;
    ssize_t         textlength = 0; /*                                       */
    ssize_t         fullrownum = 0; /*                                     */
    ssize_t         hirowborder;    /* ,                 */
    ssize_t         tbwidth;        /*                                                  */
    int             ex = 0;
    copy_style     *st;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!reg_style) {
        ifBLIN_QX0("No style");
        ERROUT(EX_DATAERR, EFAULT);
    }
    st = reg_style->style;
    if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
        ifBLIN_QX0("Illegal style %s", st->id);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (low >= 0) fullrownum = low;
    hirowborder = fullrownum + pgoblin_db_resinfo(PGOBLIN_Ntuples, pq);
    tbwidth = pgoblin_db_resinfo(PGOBLIN_Nfields, pq);
    for (ssize_t lorow = fullrownum; fullrownum < hirowborder; fullrownum++) {
        for (int colnum = 0; colnum < tbwidth; colnum++) {
            ssize_t localrownum = fullrownum - lorow;
            char   *buf;
            char   *c;
            size_t  l;
            ssize_t k;


            if  (ex < 0) goto out;
            if  (!pgoblin_db_valinfo(PGOBLIN_IsNull, pq, localrownum, colnum)) {
                l = pgoblin_db_valinfo(PGOBLIN_Length, pq, localrownum, colnum);
                c = pgoblin_db_getvalue(pq, localrownum, colnum);
            } else {
                l = 0;
                c = NULL;
            }
            if  (!(buf = malloc((c ? l * 2 : 2) + 2))) {
                ifBLIN_QW0("No mem");
                ex = -EX_OSERR;
                goto out;
            }
            if  (0 > (k = pgoblin_onecopy(buf, c, l, (colnum < (tbwidth - 1)) ? -1 : 0))) {
                ex = k;
                goto out;
            }
            buf[k] = 0;
            ex = reg_out->wri(reg_out->onu, buf, k);
            free(buf);
            if  (ex < 0) goto out;                                                                        \
            textlength += ex;                                                                                      \
    }   }
    if  (st->flags & PGOBLIN_COPE) ex = reg_out->wri(reg_out->onu, "\\.\n", 3);
out:
    return((ex < 0) ? ex : textlength);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
init(u_int32_t flags) {                                             /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
fini() {                                                            /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

pgoblin_styles PGOBLIN_MODULE =
{ 0
, "5pGoblin-" VERS
, "copy\0" VERS
, 0
, init
, fini
, parser
, dump
, clear
, get
, tbl
, NULL
, NULL
, NULL
, NULL
};
